from os import path
from queue import Queue
from sys import stderr
from threading import Thread
from time import sleep
from typing import Callable, List, NamedTuple, Optional, Sequence, Text

from psutil import AccessDenied, NoSuchProcess, Popen, Process
from psutil._pslinux import popenfile

from .progress import Output, Progress

Argv = Optional[Sequence[Text]]


def run_main(main: Callable[[], None]):
    """
    Runs the main function. Add try/catch wrappers or whatever you need here.
    That's useful in case you want to have several points to call main().

    Parameters
    ----------
    main
        Main function
    """
    return main()


class FileInfo(NamedTuple):
    """
    Current information about the file
    """

    path: Text
    size: int
    position: int


class SpyProcess:
    """
    Spying process to detect the currently open files and their current reading
    advancement.

    Notes
    -----
    There is three threads at play here:

    - The main thread, which handles the display
    - The ticking thread which drives refreshing
    - The process watching thread, which waits for the process to be done

    Both the ticking and the process threads communicate their ticks to the
    main thread through a queue. This way the main thread can easily display
    an update every second and close instantly when the process is done.
    """

    def __init__(
        self, args: Sequence[Text], period: float, output: Output, attach=Optional[int]
    ):
        self.args = args
        self.attach = attach
        self.proc: Optional[Popen] = None
        self.files_cache = {}
        self.display_ticks = Queue()
        self.process_thread = Thread(target=self.watch_process)
        self.ticks_thread = Thread(
            target=self.generate_ticks, args=(period,), daemon=True
        )
        self.progress = Progress(output)
        self.counters = {}

    def start(self):
        """
        If a PID was supplied to attach in the CLI options then use it.
        Otherwise use the provided arguments to start the command.
        """

        if self.attach is None:
            try:
                self.proc = Popen(self.args)
            except FileNotFoundError:
                stderr.write(f'Could not find command "{self.args[0]}"\n')
                exit(1)
        else:
            try:
                self.proc = Process(self.attach)
            except (AccessDenied, NoSuchProcess):
                stderr.write(
                    f"Could not attach process {self.attach}. Does it exist? "
                    f"Do you have the rights?\n"
                )
                exit(1)

    def open_files(self) -> List[popenfile]:
        """
        Returns the list of open files
        """

        return self.proc.open_files()

    def list_files(self) -> Sequence[FileInfo]:
        """
        Generates the FileInfo object of all interesting files.
        """

        files_in_use = set()
        out = []

        try:
            for f in self.open_files():
                if f.mode != "r":
                    continue

                files_in_use.add(f.path)

                if f.path not in self.files_cache:
                    self.files_cache[f.path] = path.getsize(f.path)

                if hasattr(f, "position"):
                    out.append(
                        FileInfo(
                            path=f.path,
                            size=self.files_cache[f.path],
                            position=f.position,
                        )
                    )
        except (AccessDenied, NoSuchProcess):
            pass

        for k in set(self.files_cache.keys()) - files_in_use:
            del self.files_cache[k]

        return out

    def print_progress(self):
        """
        UI display thread, looping around until the thing is done
        """

        stop = False

        try:
            while not stop:
                files = self.list_files()
                self.progress.update(files)
                stop = self.display_ticks.get()
        finally:
            self.progress.close()

    def generate_ticks(self, period: float):
        """
        Ticks into the queue every "period" second

        Parameters
        ----------
        period
            Number of seconds between two ticks
        """

        while True:
            self.display_ticks.put(False)
            sleep(period)

    def start_display(self):
        """
        Starts the threads that will tick the display
        """

        self.ticks_thread.start()
        self.process_thread.start()
        self.print_progress()

    def watch_process(self):
        """
        Waits until the process finishes and raises the tick
        """

        self.return_code()
        self.display_ticks.put(True)

    def return_code(self) -> int:
        """
        Waits for the process to finish and returns its return code
        """

        return self.proc.wait()

    def send_signal(self, sig):
        """
        Sends a signal to the child process

        Parameters
        ----------
        sig
            Unix signal
        """

        try:
            self.proc.send_signal(sig)
        except NoSuchProcess:
            pass


def positive_int(x) -> int:
    """
    Checks that the provided input is a positive integer. Used for PID
    validation in the CLI arguments.

    Parameters
    ----------
    x
        A positive integer

    Returns
    -------

    """

    x = int(x)

    if x < 0:
        raise ValueError("A positive integer is expected")

    return x



import sys
from enum import Enum
from fcntl import ioctl
from io import BufferedWriter
from os import getenv, path
from signal import SIGWINCH, signal
from struct import unpack
from termios import TIOCGWINSZ
from typing import TYPE_CHECKING, Callable, Dict, NamedTuple, Sequence, Text, TextIO

from tqdm import tqdm

if TYPE_CHECKING:
    from .utils import FileInfo


SINK = open("/dev/null", "r+")


def shorten_file_path(max_length: int, file_path: Text):
    """
    Applies several strategies to try and condense the file path

    Notes
    -----
    The strategies are the following:

    1. If the path starts with $HOME, replace the beginning with ~/
    2. Remove as many parent dirs as needed to fit
    3. If only the file name is left, abbreviate it. However, try to keep the
       extension visible in doing so.

    I'm guessing that on Windows this would be kind of broken but for now there
    is no Windows compatibility.

    Parameters
    ----------
    max_length
        The output value will not be longer than this

    file_path
        The file path you want to shorten

    Returns
    -------
    A shorted file path. By example "/home/foo/bar.txt" becomes "~/bar.txt"
    (if your home is "/home/foo").
    """

    home = getenv("HOME", None)

    if home:
        home = home.rstrip("/")

        if file_path.startswith(home):
            file_path = "~" + file_path[len(home) :]

    if len(file_path) <= max_length:
        return file_path

    parts = file_path.split("/")

    for i in range(1, len(parts) - 1):
        file_path = path.join(*parts[i:])

        if len(file_path) <= max_length:
            return file_path

    name, ext = path.splitext(file_path)
    max_name_length = max_length - len(ext)

    if max_name_length < 5:
        return f"{file_path[0:max_length-3]}..."
    else:
        return f"{name[0:max_name_length-2]}..{ext}"


class Output(Enum):
    """
    Used to restrict potential values of output files from the argument
    parsing.
    """

    stderr = "stderr"
    stdout = "stdout"


class TerminalSize(NamedTuple):
    """
    Contains the output of what ioctl tells us on the terminal's size
    """

    rows: int
    cols: int
    x_pixel: int
    y_pixel: int

    @classmethod
    def from_bytes(cls, b: bytes) -> "TerminalSize":
        """
        Converts the output of ioctl into a workable TerminalSize

        Parameters
        ----------
        b
            Byte values

        Returns
        -------
        Parsed value
        """
        return cls(*unpack("hhhh", b))


class FileProgress(NamedTuple):
    """
    Information about the current progress of file
    """

    info: "FileInfo"
    tqdm: tqdm


class BottomBox:
    """
    Utility class to display something at the bottom of the screen while the
    stdout can still print things.

    Notes
    -----
    This instructs the terminal to create a scroll window above the text in the
    bottom. When an update happens, bottom lines are manually drawn. We also
    manage the cursor so that when stdout prints something it is at the bottom
    of the window it scrolls up and is not drawn downstairs.

    This happens on stderr, this way the output of stdout can be piped.

    The renderer will either get called on updates (aka when you call update())
    either when the terminal is resized.

    Also, there is an inner buffer before writing out things to make sure that
    all characters are written at the same time and avoid messing things up.
    """

    def __init__(self, renderer: Callable[[int], Sequence[Text]], output: Output):
        """
        Constructor.

        Parameters
        ----------
        renderer
            A callable that will be called when a drawing is requested
        """

        self.last_height = 0
        self.stdout: TextIO = getattr(sys, output.value)
        self.renderer = renderer
        # noinspection PyTypeChecker
        self.buffer = BufferedWriter(self.stdout.buffer, buffer_size=1_000_000)

        signal(SIGWINCH, lambda _, __: self.update())

    def _write(self, s: bytes):
        """
        Store bytes into the buffer

        Parameters
        ----------
        s
            Bytes to be flushed later
        """

        self.buffer.write(s)

    def _flush(self):
        """
        Flushes the bytes to the output
        """

        self.buffer.flush()
        self.stdout.buffer.flush()

    def _save_cursor(self) -> None:
        """
        ANSI save of the cursor position
        """

        self._write(b"\0337")

    def _restore_cursor(self) -> None:
        """
        ANSI restore of the cursor position
        """

        self._write(b"\0338")

    def _set_scroll_region(self, rows: int) -> None:
        """
        ANSI definition of the scroll area
        Parameters
        ----------
        rows
            Number of rows in the scroll area
        """

        self._write(f"\033[0;{rows}r".encode())

    def _move_cursor_to_row(self, row: int) -> None:
        """
        Moves the cursor to that row

        Parameters
        ----------
        row
            1-indexed row number
        """

        self._write(f"\033[{row};1f".encode())

    def _move_cursor_up(self) -> None:
        """
        Moves the cursor to the row up
        """

        self._write(b"\033[1A")

    def _clear_row(self) -> None:
        """
        Clears the current row
        """

        self._write(b"\033[K")

    def _get_terminal_size(self) -> TerminalSize:
        """
        Inquires the size of the terminal and returns it
        """

        return TerminalSize.from_bytes(
            ioctl(self.stdout.fileno(), TIOCGWINSZ, "\0" * 8)
        )

    def _adjust_for_height(self, new_height) -> None:
        """
        Makes sure that there is white space enough at the bottom of the output
        in order for the bottom box to be drawn.
        """

        for _ in range(self.last_height, new_height):
            self._write(b"\n")

        for _ in range(self.last_height, new_height):
            self._move_cursor_up()

        self.last_height = new_height

    def _clear_screen_below_cursor(self) -> None:
        """
        Clears everything below the cursor using dedicated ANSI code
        """

        self._write(b"\033[J")

    def update(self) -> None:
        """
        Updates the content of the bottom box
        """

        size = self._get_terminal_size()
        lines = self.renderer(size.cols)

        self._adjust_for_height(len(lines))
        self._save_cursor()
        self._set_scroll_region(size.rows - len(lines))

        start = size.rows - len(lines) + 1

        for i, line in enumerate(lines):
            self._move_cursor_to_row(start + i)
            self._clear_row()
            self._write(line.encode())

        self._restore_cursor()
        self._flush()

    def cleanup(self):
        """
        When done, restores the scrolling idea and the configuration of the
        terminal (as much as possible).
        """

        size = self._get_terminal_size()
        self._save_cursor()
        self._set_scroll_region(size.rows)
        self._restore_cursor()
        self._clear_screen_below_cursor()
        self._flush()


class Progress:
    """
    Displays the progress of the files, based on tqdm but with custom rendering
    """

    MIN_FILE_WIDTH = 10
    FILE_WIDTH = 0.25

    def __init__(self, output: Output):
        self.box = BottomBox(self._render, output)
        self.cnt = 0
        self.files: Dict[Text, FileProgress] = {}

    def _render(self, width: int) -> Sequence[Text]:
        """
        Renders all the the lines to be displayed.

        Parameters
        ----------
        width
            Terminal width

        Returns
        -------
        List of lines to be displayed
        """

        out = []

        file_width = int(max([self.MIN_FILE_WIDTH, self.FILE_WIDTH * width]))

        for file, progress in self.files.items():
            progress.tqdm.ncols = width
            progress.tqdm.update(progress.info.position - progress.tqdm.n)
            progress.tqdm.desc = shorten_file_path(file_width, file)
            out.append(repr(progress.tqdm))

        return out

    def update(self, files: Sequence["FileInfo"]) -> None:
        """
        Re-generates the progress bars from the files list.

        Notes
        -----
        We use tqdm here, however we don't want it to be handling things on
        itself. For this reason we set it to output its content to /dev/null
        and we just re-render it using the repr() method.

        Parameters
        ----------
        files
            Current status of files being read. The proper diff since last
            update will be made automatically.
        """

        present = set()

        for file in files:
            present.add(file.path)

            if file.path not in self.files:
                self.files[file.path] = FileProgress(
                    file,
                    tqdm(
                        desc=file.path,
                        total=file.size,
                        unit="o",
                        smoothing=0.1,
                        unit_divisor=1024,
                        unit_scale=True,
                        file=SINK,
                    ),
                )
            else:
                self.files[file.path] = self.files[file.path]._replace(info=file)

        for absent in set(self.files.keys()) - present:
            del self.files[absent]

        self.box.update()

    def close(self):
        """
        Triggers screen cleanup
        """

        self.box.cleanup()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['spybar']

package_data = \
{'': ['*']}

install_requires = \
['psutil>=5.6.7,<6.0.0', 'tqdm>=4.41.1,<5.0.0']

entry_points = \
{'console_scripts': ['spybar = spybar.__main__:run']}

setup_kwargs = {
    'name': 'spybar',
    'version': '0.1.0',
    'description': 'Adds progress bars to any file-reading tool (gzip, md5sum, etc)',
    'long_description': "# Spybar\n\nSpybar is a Linux CLI utility that adds a progress bar to any tool reading a\nfile. By example, it can create a progress bar for `gzip`, `md5sum` and other\ntraditional utilities which don't display their progress.\n\nSimply prefix any command with `spybar`.\n\n```\nspybar gzip big_dump.sql\n```\n\nOr, if the process is already running, attach using its PID. Suppose that you\nwant to attach to process 42:\n\n```\nspybar -a 42\n```\n\n## FAQ\n\n### Can you pipe it?\n\nYes you can, the progress bar happens on `stderr`\n\n```\nspybar gzip -c big_dump.sql > big_dump.sql.gz\n```\n\n### Can it support Win/OSX?\n\nUnfortunately there is no known way to do so. This utility relies on `psutil`'s\n[`Process.open_files()`](https://psutil.readthedocs.io/en/latest/index.html#psutil.Process.open_files)\nabstraction so whenever this abstraction gives you the `position` for other\nplatforms than Linux then Spybar will automatically work on those.\n\n(By automatically I mean there is maybe a few adjustments needed on \nUnix-specific assumptions but it's most likely a just few lines to change).\n\n### How does it work?\n\nIf you navigate to the `/proc` file system you will see that for each process\nyou not only get the list of open files (in `/proc/XX/fd`) but also \nmeta-information about those files (in `/proc/XX/fdinfo`).\n\nThis way Spybar will look for files open in read mode by your process and then\nlook at the current position of the file pointer, which once compared to the\nfile size gives you the relative progress.\n\n### Does it always work?\n\nOf course not, but the use case of binaries reading a file from the beginning\nto the end is fairly common. Who never waited in front of a `gzip`, `xz`, `tar`\nor `md5sum` wondering if they should go have a coffee or if it's just 2 seconds\nmore?\n\nIn the end this is just guessing but it works in many situations.\n\n## Thanks\n\nI would like to thank:\n\n- Whomever put this feature in the Linux kernel\n- The `psutil` maintainers\n- The `tqdm` maintainers\n- The `poetry` maintainers\n- The `pytest` maintainers\n- All open-source contributors thanks to whom this software was easy to write\n\n## License\n\nThis software is released under the terms of the [WTFPL](./LICENSE).\n",
    'author': 'Rémy Sanchez',
    'author_email': 'remy.sanchez@hyperthese.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import signal
from argparse import REMAINDER, ArgumentParser, Namespace

from .progress import Output
from .utils import Argv, SpyProcess, positive_int, run_main


def parse_args(argv: Argv = None) -> Namespace:
    """
    Parses arguments by default from CLI but if argv is specified parses this
    instead of the default arguments.

    Parameters
    ----------
    argv
        Optional list of arguments to parse. The system argv is used by
        default.

    Returns
    -------
    The parsed Namespace
    """

    parser = ArgumentParser(
        description="Adds a progress bar to any file-reading command"
    )

    parser.add_argument(
        "-a",
        "--attach",
        type=positive_int,
        help=(
            "Instead of starting a process, attaches to an already-running "
            "process. Specify the PID here."
        ),
    )
    parser.add_argument(
        "-p",
        "--refresh-period",
        type=float,
        help=(
            "Period in seconds between two refreshes of the progress bars "
            "(default: 1s)"
        ),
        default="1",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Output,
        help='Output the bar to "stderr" or "stdout" (default: "stderr")',
        default=Output.stderr,
    )
    parser.add_argument(
        "command_arg",
        nargs=REMAINDER,
        help=(
            "After the options, just type your regular command, by example "
            '"spybar gzip big_dump.sql"'
        ),
    )

    ns = parser.parse_args(argv)

    if ns.attach is None and not ns.command_arg:
        parser.error(
            "You need to specify at least a PID to attach (-a option) or a "
            'command to run. By example "spybar gzip bigdump.sql"'
        )
    elif ns.attach is not None and ns.command_arg:
        parser.error(
            "You cannot specify a PID to attach and a "
            "command to run at the same time"
        )

    return ns


def main(argv: Argv = None):
    """
    Runs the process spy.

    Notes
    -----
    Most things happens in SpyProcess, however this function shows you the
    regular call order.
    """

    args = parse_args(argv)

    sp = SpyProcess(
        args=args.command_arg,
        period=args.refresh_period,
        output=args.output,
        attach=args.attach,
    )
    sp.start()

    try:
        sp.start_display()
    except KeyboardInterrupt:
        sp.send_signal(signal.SIGINT)
    except SystemExit:
        sp.send_signal(signal.SIGTERM)
    finally:
        ret = sp.return_code()

    if ret is not None:
        exit(ret)


def run():
    """
    Convenient function which can be called from a binary installed by the
    setup script.
    """

    run_main(main)


if __name__ == "__main__":
    run()
